from asgi_aiogram.aliases import Receiver

async def read_body(receive: Receiver) -> bytes:
    body = b''
    more_body = True

    while more_body:
        message = await receive()
        body += message.get('body', b'')
        more_body = message.get('more_body', False)

    return body


from json import loads, dumps
from typing import Callable

from aiogram.dispatcher.event.handler import CallableObject

from asgi_aiogram.aliases import Receiver, Sender
from asgi_aiogram.http_request import Request
from asgi_aiogram.http_responses import HttpResponse
from asgi_aiogram.strategy import BaseStrategy
from asgi_aiogram.types import ScopeType


class HttpStrategy(BaseStrategy):
    kwargs: dict

    async def startup(self, kwargs: dict):
        self.kwargs = dict(kwargs)

    async def handle(self, scope: ScopeType, receive: Receiver, send: Sender) -> bytes | None:
        response = await self._handler.call(
            request=Request(scope=scope, receive=receive, send=send, strategy=self),
            **self.kwargs
        )
        if isinstance(response, HttpResponse):
            await response(scope=scope, receive=receive, send=send)
        return

    def verify_path(self, scope: ScopeType) -> bool:
        if scope["method"] != self._method:
            return False
        if scope["path"] != self._path:
            return False
        return True

    def __init__(self, method: str, path: str, handler: Callable, loads: Callable = loads, dumps: Callable = dumps):
        super().__init__(loads=loads, dumps=dumps)
        self._path = path
        self._method = method.upper()
        self._handler = CallableObject(handler)


from typing import TypedDict, Literal, NotRequired, Iterable

class ScopeType(TypedDict):
    type: Literal['http', 'lifespan']
    scheme: str
    root_path: str
    server: tuple[str, int]
    http_version: str
    method: str
    path: str
    headers: list[tuple[bytes, bytes]]

class SendEventType(TypedDict):
    type: Literal[
        "http.response.start",
        "http.response.body",
        "lifespan.startup.complete",
        "lifespan.startup.failed",
        "lifespan.shutdown.complete",
        "lifespan.shutdown.failed",
    ]
    status: NotRequired[int]
    headers: NotRequired[Iterable[tuple[bytes, bytes]]]
    body: NotRequired[bytes]
    more_body: NotRequired[bool]

class ReceiveEventType(TypedDict):
    type: str
    body: NotRequired[bytes]
    more_body: NotRequired[bool]


from asyncio import gather
from typing import Any, Sequence

from aiogram import Bot, Dispatcher

from asgi_aiogram.path_normalization import split_path_by_token_placeholder
from asgi_aiogram.strategy.base import BaseBotStrategy
from asgi_aiogram.types import ScopeType


class TokenBasedStrategy(BaseBotStrategy):
    def __init__(self, path: str, dispatcher: Dispatcher, bot_settings: dict[str, Any]):
        super().__init__()
        self._path_prefix, self._slice, self._path_postfix = split_path_by_token_placeholder(path)
        self._bots: dict[str, Bot] = {}
        self._bot_settings = bot_settings
        self.dispatcher = dispatcher

    def verify_path(self, scope: ScopeType) -> bool:
        if scope["method"] != "POST":
            return False
        if not scope["path"].startswith(self._path_prefix):
            return False
        if not scope["path"].endswith(self._path_postfix):
            return False
        return True

    async def resolve_bot(self, scope: ScopeType) -> Bot | None:
        token = scope["path"][self._slice]
        if token not in self._bots:
            self._bots[token] = Bot(token=token, **self._bot_settings)
        return self._bots[token]

    async def shutdown(self, kwargs: dict):
        await super().shutdown(kwargs=kwargs)
        await gather(*(bot.session.close() for bot in self._bots.values()))

    @property
    def bots(self) -> Sequence[Bot]:
        return list(self._bots.values())



PLACEHOLDER_MISSING_ERROR_MESSAGE: str = "Path does not contain token placeholder {token_placeholder}"

class PlaceholderMissingException(Exception):
    def __init__(self, token_placeholder: str):
        super().__init__(PLACEHOLDER_MISSING_ERROR_MESSAGE.format(
            token_placeholder=token_placeholder
        ))


from logging import getLogger
from typing import Any

from asgi_aiogram.aliases import Receiver, Sender
from asgi_aiogram.responses import not_found
from asgi_aiogram.strategy import BaseStrategy
from asgi_aiogram.types import ScopeType


class ASGIAiogram:
    def __init__(self,
        *strategies: BaseStrategy,
        **kwargs: Any
    ):
        self.strategies = strategies
        self.kwargs = kwargs
        self.logger = getLogger("asgi_aiogram")
        self.resolve_chance = {}

    def resolve_strategy(self, scope: ScopeType) -> BaseStrategy | None:
        key = (scope["method"], scope["path"])
        if key not in self.resolve_chance:
            for strategy in self.strategies:
                if strategy.verify_path(scope=scope):
                    self.resolve_chance[key] = strategy
                    return strategy
            return None
        return self.resolve_chance[key]


    async def lifespan(self, scope: ScopeType, receive: Receiver, send: Sender):
        while True:
            message = await receive()
            if message['type'] == 'lifespan.startup':
                try:
                    for strategy in self.strategies:
                        await strategy.startup(kwargs=self.kwargs)
                except Exception as e:
                    self.logger.error(e)
                    await send({'type': 'lifespan.startup.failed'})
                else:
                    await send({'type': 'lifespan.startup.complete'})
            elif message['type'] == 'lifespan.shutdown':
                try:
                    for strategy in self.strategies:
                        await strategy.shutdown(kwargs=self.kwargs)
                except Exception as e:
                    self.logger.error(e)
                    await send({'type': 'lifespan.shutdown.failed'})
                    return
                else:
                    await send({'type': 'lifespan.shutdown.complete'})
                    return

            else:
                self.logger.warning("unknown lifespan type: %s", message['type'])

    async def http(self, scope: ScopeType, receive: Receiver, send: Sender):
        strategy = self.resolve_strategy(scope=scope)
        if strategy is None:
            self.logger.warning("unknown request. method: %s path: %s", scope["method"], scope['path'])
            await not_found(send=send)
            return
        return await strategy.handle(scope=scope, receive=receive, send=send)



    async def __call__(self, scope: ScopeType, receive: Receiver, send: Sender):
        if scope['type'] == 'http':
            return await self.http(scope=scope, receive=receive, send=send)
        if scope['type'] == 'lifespan':
            return await self.lifespan(scope=scope, receive=receive, send=send)
        self.logger.warning("unsupported event type:", scope['type'])


from typing import Callable, TypeVar, Union

from asgi_aiogram.aliases import Receiver, Sender
from asgi_aiogram.asgi import read_body
from asgi_aiogram.strategy import BaseStrategy
from asgi_aiogram.types import ScopeType

NO_HEADER = object()
T = TypeVar('T')

class Headers:
    def __init__(self, scope: ScopeType):
        self.scope = scope
        self._chance_one: dict[str, Union[str, NO_HEADER]] = {}
        self._chance_all: dict[str, list[str]] = {}

    def _get_one(self, key: str | bytes) -> Union[str, NO_HEADER]:
        if key not in self._chance_one:
            if isinstance(key, str):
                b_key = key.encode()
            else:
                b_key = key
            for h_name, h_value in self.scope["headers"]:
                if h_name == b_key:
                    self._chance_one[key] = h_value.decode()
                    break
            self._chance_one[key] = NO_HEADER
        return self._chance_one[key]

    def _get_all(self, key: str | bytes) -> list[str]:
        if key not in self._chance_all:
            self._chance_all[key] = items = []
            if isinstance(key, str):
                b_key = key.encode()
            else:
                b_key = key
            for h_name, h_value in self.scope["headers"]:
                if h_name == b_key:
                    items.append(h_value.decode())
        return self._chance_all[key]

    def __getitem__(self, key: str):
        header = self._get_one(key=key)
        if header is NO_HEADER:
            raise KeyError(key)
        return header

    def get(self, key: str | bytes, default: T = None) -> str | T:
        header = self._get_one(key=key)
        if header is NO_HEADER:
            return default
        return header

    def get_all(self, key: str | bytes) -> list[str] | list[str]:
        return self._get_all(key=key)

    def __setitem__(self, key, value):
        raise Exception("Readonly")

    @property
    def raw(self):
        return self.scope["headers"]



class Request:
    def __init__(self, scope: ScopeType, receive: Receiver, send: Sender, strategy: BaseStrategy):
        self._scope = scope
        self._receive = receive
        self._send = send
        self._strategy = strategy
        self._body = None
        self._json = None
        self._text = None
        self._headers = None

    async def body(self) -> bytes:
        if self._body is None:
            self._body = await read_body(self._receive)
        return self._body

    async def json(self, loads: Callable | None = None) -> dict:
        if self._json is None:
            loads = loads or self._strategy.loads
            self._json = loads(await self.body())
        return self._json

    async def text(self, encoding: str = "utf-8") -> str:
        if self._text is None:
            self._text = (await self.body()).decode(encoding=encoding)
        return self._text

    @property
    def headers(self) -> Headers:
        if self._headers is None:
            self._headers = Headers(self._scope)
        return self._headers




from typing import Callable, Awaitable, TypeAlias

from asgi_aiogram.types import SendEventType, ReceiveEventType

Sender: TypeAlias = Callable[[SendEventType], Awaitable[None]]

Receiver: TypeAlias = Callable[[], Awaitable[ReceiveEventType]]


from abc import abstractmethod
from io import BytesIO
from json import dumps
from typing import Any, Callable

from asgi_aiogram.aliases import Receiver, Sender
from asgi_aiogram.types import ScopeType


def _prepare_header_value(self, value: Any) -> bytes:
    if isinstance(value, bytes):
        return value
    if isinstance(value, (list, tuple, set)):
        return b", ".join(_prepare_header_value(i) for i in value)
    return str(value).encode()


def _prepare_header_key(self, key: str) -> bytes:
    return key.lower().encode()


class HttpResponse:
    _prepare_header_value: Callable[["HttpResponse", Any], bytes] = _prepare_header_value
    _prepare_header_key: Callable[["HttpResponse", str], bytes] = _prepare_header_key
    _headers: list[tuple[bytes, bytes]]

    def _prepare_headers(self, headers: dict[str, Any]) -> list[tuple[bytes, bytes]]:
        return [
            (self._prepare_header_key(key), self._prepare_header_value(value))
            for key, value in headers.items()
        ]

    def __init__(self, status_code: int = 200, headers: dict[str, str] | None = None) -> None:
        self._status_code = status_code
        if headers is None:
            self._headers = []
        else:
            self._headers = self._prepare_headers(headers=headers)

    def add_header(self, key: bytes, value: bytes, rewrite: bool = True) -> None:
        for header_item in self._headers:
            if header_item[0] == key:
                if rewrite:
                    self._headers.remove(header_item)
                else:
                    return
        self._headers.append((key, value))

    def set_content_length(self, content_length: int | bytes, rewrite: bool = True) -> None:
        self.add_header(key=b'content-length', value=self._prepare_header_value(content_length), rewrite=rewrite)

    def set_content_type(self, content_type: str | bytes, rewrite: bool = True) -> None:
        self.add_header(key=b'content-type', value=self._prepare_header_value(content_type), rewrite=rewrite)

    @abstractmethod
    async def __call__(self, scope: ScopeType, receive: Receiver, send: Sender):
        pass


class EmptyResponse(HttpResponse):

    async def __call__(self, scope: ScopeType, receive: Receiver, send: Sender):
        await send({
            'type': 'http.response.start',
            'status': self._status_code,
            'headers': self._headers,
        })
        await send({
            'type': 'http.response.body',
        })

class BytesResponse(HttpResponse):
    async def __call__(self, scope: ScopeType, receive: Receiver, send: Sender):
        await send({
            'type': 'http.response.start',
            'status': self._status_code,
            'headers': self._headers,
        })
        while True:
            chunk = self._body.read(self._chunk_size)
            if chunk:
                await send({
                    'type': 'http.response.body',
                    'body': chunk,
                    'more_body': self._size != self._body.tell(),
                })
            break

    def __init__(
            self,
            body: bytes | BytesIO,
            chunk_size: int = 65536,
            status_code: int = 200,
            headers: dict[str, str] | None = None
    ) -> None:
        super().__init__(status_code=status_code, headers=headers)
        if isinstance(body, bytes):
            body = BytesIO(body)
        self._body = body
        self._size = body.getbuffer().nbytes
        self._chunk_size = chunk_size
        self.set_content_type(b"application/octet-stream", False)
        self.set_content_length(self._size)

class JsonResponse(BytesResponse):
    def __init__(
            self,
            body: dict | list | str | bytes,
            dumper: Callable[[str], bytes] = dumps,
            chunk_size: int = 65536,
            status_code: int = 200,
            headers: dict[str, str] | None = None
    ) -> None:
        body = dumper(body)
        if isinstance(body, str):
            body = body.encode()
        super().__init__(body=body, status_code=status_code, headers=headers, chunk_size=chunk_size)
        self.set_content_type(b"application/json")



from aiohttp.formdata import FormData

from asgi_aiogram.aliases import Sender


async def answer(send: Sender, form: FormData) -> None:
    data = form()
    body = data.decode().encode(encoding='utf-8')
    await send({
        'type': 'http.response.start',
        'status': 200,
        'headers': [
            (b'content-type', data.content_type.encode()),
            (b'content-length', str(data.size or len(body)).encode())
        ]
    })
    await send({
        'type': 'http.response.body',
        'body': body,
    })

async def _send_code(send: Sender, code: int) -> None:
    await send({
        'type': 'http.response.start',
        'status': code,
    })
    await send({
        'type': 'http.response.body',
    })

async def ok(send: Sender) -> None:
    await _send_code(send, 200)

async def error(send: Sender) -> None:
    await _send_code(send, 500)

async def not_found(send: Sender) -> None:
    await _send_code(send, 404)


from abc import ABC, abstractmethod
from asyncio import Task, create_task
from collections.abc import Callable
from logging import Logger, getLogger
from typing import Sequence
from json import loads, dumps

from aiogram import Bot, Dispatcher
from aiogram.types import Update

from asgi_aiogram.aliases import Receiver, Sender
from asgi_aiogram.asgi import read_body
from asgi_aiogram.responses import not_found, error, ok
from asgi_aiogram.types import ScopeType


class BaseStrategy(ABC):
    logger: Logger
    loads: Callable = loads
    dumps: Callable = dumps

    def __init__(self, loads: Callable = loads, dumps: Callable = dumps):
        self.logger = getLogger(self.__class__.__name__)
        self.loads = loads
        self.dumps = dumps

    @abstractmethod
    def verify_path(self, scope: ScopeType) -> bool:
        pass

    async def startup(self, kwargs: dict):
        pass

    async def shutdown(self, kwargs: dict):
        pass

    @abstractmethod
    async def handle(self, scope: ScopeType, receive: Receiver, send: Sender) -> bytes | None:
        pass

class BaseBotStrategy(BaseStrategy):
    dispatcher: Dispatcher
    kwargs: dict
    task_list: set[Task]
    handle_as_tasks: bool = True

    def __init__(self, loads: Callable = loads, dumps: Callable = dumps):
        super().__init__(loads=loads, dumps=dumps)
        self.task_list = set()

    @property
    def bots(self) -> Sequence[Bot]:
        return []

    @property
    def bot(self) -> Bot | None:
        return

    async def startup(self, kwargs: dict):
        self.kwargs = dict(kwargs)
        self.kwargs.pop("bot", None)
        self.kwargs.pop("dispatcher", None)
        await self.dispatcher.emit_startup(
            **self.dispatcher.workflow_data,
            **self.kwargs,
            bots=self.bots,
            bot=self.bot,
            dispatcher=self.dispatcher,
        )

    async def shutdown(self, kwargs: dict):
        self.kwargs = dict(kwargs)
        self.kwargs.pop("bot", None)
        self.kwargs.pop("dispatcher", None)
        await self.dispatcher.emit_shutdown(
            **self.dispatcher.workflow_data,
            **self.kwargs,
            bots=self.bots,
            bot=self.bot,
            dispatcher=self.dispatcher,
        )

    @abstractmethod
    async def resolve_bot(self, scope: ScopeType) -> Bot | None:
        pass

    async def handle(self, scope: ScopeType, receive: Receiver, send: Sender) -> bytes | None:
        bot = await self.resolve_bot(scope=scope)
        if bot is None:
            self.logger.warning("unknown bot for: %s", scope['path'])
            await not_found(send=send)
            return
        try:
            cor = self.dispatcher.feed_update(
                bot=bot,
                update=Update.model_validate_json(await read_body(receive)),
                **self.kwargs,
            )
            if self.handle_as_tasks:
                handle_update_task = create_task(cor)
                self.task_list.add(handle_update_task)
                handle_update_task.add_done_callback(self.task_list.discard)
            else:
                await cor
                # if isinstance(response, TelegramMethod):
                #     form = bot.session.build_form_data(bot, response)
                #     form.add_field(
                #         name="method",
                #         value=response.__api_method__
                #     )
                #     await answer(send, form)
            await ok(send=send)
            return
        except Exception as e:
            self.logger.error(e)
            await error(send=send)



from typing import Sequence

from aiogram import Bot, Dispatcher

from asgi_aiogram.strategy.base import BaseBotStrategy
from asgi_aiogram.types import ScopeType


class SingleStrategy(BaseBotStrategy):
    def __init__(self, path: str, bot: Bot, dispatcher: Dispatcher):
        super().__init__()
        self._path = path
        self._bot = bot
        self.dispatcher = dispatcher

    def verify_path(self, scope: ScopeType) -> bool:
        return self._path == scope['path'] and scope["method"] == "POST"

    async def resolve_bot(self, scope: ScopeType) -> Bot | None:
        return self._bot

    async def shutdown(self, kwargs: dict):
        await super().shutdown(kwargs=kwargs)
        await self._bot.session.close()

    @property
    def bots(self) -> Sequence[Bot]:
        return [self._bot]

    @property
    def bot(self) -> Bot | None:
        return self._bot


from asgi_aiogram.exceptions import PlaceholderMissingException

TOKEN_PLACEHOLDER = "{bot_token}"
TOKEN_PLACEHOLDER_LIGHT = 11


def split_path_by_token_placeholder(path: str) -> tuple[str, slice, str]:
    if TOKEN_PLACEHOLDER not in path:
        raise PlaceholderMissingException(TOKEN_PLACEHOLDER)
    token_start_index = path.index(TOKEN_PLACEHOLDER)
    token_end_index = token_start_index + TOKEN_PLACEHOLDER_LIGHT

    path_prefix = path[:token_start_index]
    path_suffix = path[token_end_index:]

    slice_end = token_end_index - len(path) or None

    return path_prefix, slice(token_start_index, slice_end), path_suffix
